/*      
   Tsunami for KallistiOS ##version##

   list.h

   (c)2002 Dan Potter

   $Id: list.h,v 1.2 2002/09/10 05:48:22 bardtx Exp $
*/

#ifndef __TSUNAMI_LIST_H
#define __TSUNAMI_LIST_H

/// \file A basic linked-list class. The only real constraint here
/// is that the type T must be a subclass of RefCnt or something that
/// provides similar methods. This cheats a little by implementing
/// all the core functionality via the BSD queue macros, but it provides
/// a much nicer C++ interface.

#include <assert.h>
#include <string.h>
#include <sys/queue.h>
#include "refcnt.h"

template <class T>
class List;

template <class T>
struct ListNode {
public:
	ListNode<T>(T * d)
		: m_data(d), m_inlist(false)
	{ }

	virtual ~ListNode<T>() {
		assert( !m_inlist );
	}

	/// Get the next list item
	ListNode<T> * getNext() const {
		assert( m_inlist );
		if ( !m_inlist )
			return NULL;

		return LIST_NEXT(this, m_listptr);
	}

	/// Get this list item's data
	T * getData() const { return m_data; }

	/// Perform an operation on this list item's data
	T * operator->() const { return (T*)m_data; }

	/// Remove this item from the list (doesn't delete it)
	void remove() {
		assert( m_inlist );
		if (!m_inlist) return;

		LIST_REMOVE(this, m_listptr);
		m_inlist = false;
	}

	/// Insert another element before us
	void insertBefore(ListNode<T> * other) {
		assert( !m_inlist );
		if (m_inlist) return;

		LIST_INSERT_BEFORE(other, this, m_listptr);
		m_inlist = true;
	}

	/// Insert another element after us
	void insertAfter(ListNode<T> * other) {
		assert( !m_inlist );
		if (m_inlist) return;

		LIST_INSERT_AFTER(other, this, m_listptr);
		m_inlist = true;
	}

private:
	RefPtr<T>			m_data;
	bool				m_inlist;

	friend class List<T>;
	LIST_ENTRY(ListNode<T>)	m_listptr;
};


template <class T>
class List {
public:
	List() {
		memset(&m_list, 0, sizeof(m_list));
	}
	
	virtual ~List() {
		ListNode<T> * n, * t;

		t = LIST_FIRST(&m_list);
		while (t) {
			n = t->getNext();
			t->m_inlist = false;
			delete t;
			t = n;
		}
	}

	/// Insert an item at the head of the list. A reference will be added.
	void insertHead(T * item) {
		ListNode<T> * ni = new ListNode<T>(item);
		LIST_INSERT_HEAD(&m_list, ni, m_listptr);
		ni->m_inlist = true;
	}

	/// Delete the named item. Return true if we actually had it. Its
	/// reference will be removed if it is found.
	bool del(T * item) {
		ListNode<T> * n;
		
		LIST_FOREACH(n, &m_list, m_listptr) {
			if (n->getData() == item) {
				n->remove();
				delete n;
				return true;
			}
		}

		return false;
	}

	/// Return the element at the head of the list
	ListNode<T>	* getHead() const {
		return LIST_FIRST(&m_list);
	}

private:
	LIST_HEAD(/*anon*/, ListNode<T>) m_list;
};

#endif	// __TSUNAMI_LIST_H
